// Copyright (C) 2004-2007, Software Development Laboratories, "Fish" (David B. Trout)
//////////////////////////////////////////////////////////////////////////////////////////
//
//      ftape.cpp  --  Product mainline
//
//////////////////////////////////////////////////////////////////////////////////////////
//
//  Change History:
//
//  07/16/04    1.0.0   Created.
//  07/23/04    1.0.1   1) Don't close tape file if GetTapeStatus or SetTapeParameters
//                      returns an "error", 2) Spin off a worker thread to monitor when
//                      media is loaded/unloaded.
//  07/27/04    1.1.2   w32_stape_update_status: Don't skip retrieval of Drive Parms
//                      GetTapeParameters(GET_TAPE_DRIVE_INFORMATION) if tape simply
//                      isn't loaded. (media status should not affect drive status!)
//  07/27/04    1.1.3   MonitorMediaMountsThread: change GetTapeStatus loop to ignore
//                      ERROR_MEDIA_CHANGED instead of ERROR_BUS_RESET! (bus resets
//                      ARE significant, whereas media changes are not. The only media
//                      status we're really interested in is ERROR_NO_MEDIA_IN_DRIVE.
//  08/10/04    1.2.0   Added 'readpos' and 'locate' commands.
//  09/02/04    1.3.0   Added 'ascii', 'ebcdic' and 'dump' commands.
//  10/01/04    1.4.0   Added 'erase' command. (Data Security Erase)
//  12/24/04    1.4.0   Added "offset.length" argument to 'dump' command.
//  08/05/06    1.4.1   Fix tape-mount-monitoring-thread bug (MonitorMediaMountsThread)
//                      causing rougue drive status queries (i.e. when not asked to).
//  08/05/06    1.5.0   Status reporting enhancements related to variable blocksize i/o.
//  08/05/06    1.5.0   Change all functions with an "immediate / non-immediate" option
//                      to instead use whatever mode the tape drive happens to support.
//  08/23/06    1.5.1   Fix bad event test in MonitorMediaMountsThread causing hang.
//  02/04/07    1.6.0   VS2005 + x64
//  02/04/07    1.6.0   Added Ctrl+C handler to allow aborting of 'dump' command.
//  02/24/07    1.6.1   Quick fix to Ctrl+C handling to prevent loop if not 'dump'ing.
//
//////////////////////////////////////////////////////////////////////////////////////////

//                           ** Win9x is NOT supported **

#include "stdafx.h"
#include "ftape.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//////////////////////////////////////////////////////////////////////////////////////////

#define  DEFAULT_TAPE_DEVICE     "Tape0"
#define  TAPE_DEVICE_PREFIX      "\\\\.\\"
#define  RANDOM_BLKSIZE_LOW      (12)           // (anything less is noise)
#define  RANDOM_BLKSIZE_HIGH     (0xFFFF)       // (64K - 1)
#define  HEXDUMP_BYTES           (256)

//////////////////////////////////////////////////////////////////////////////////////////
// The one and only application object

CWinApp theApp;

using namespace std;

//////////////////////////////////////////////////////////////////////////////////////////

void Copyright()
{
    _tprintf( _T( "\n" ) );
    _tprintf( _T( PRODUCT ", v" VERSION_STR "\n" ) );
    _tprintf( _T( COPYRIGHT ", " COMPANY "\n" ) );
    _tprintf( _T( PRODUCT_URL " (" PRODUCT_EMAIL ")\n\n" ) );
}

//////////////////////////////////////////////////////////////////////////////////////////

void usage()
{
    _tprintf( _T( "Usage:\n\n" ) );

    _tprintf( _T( "   ftape  [-f filename]  function  [count or value]\n\n" ) );

    _tprintf( _T( "Where:\n\n" ) );

    _tprintf( _T( "   filename    name of tape device  (default=\"Tape0\")\n" ) );
    _tprintf( _T( "   function    (see below)\n" ) );
    _tprintf( _T( "   count       how many times to do it (if supported)\n\n" ) );

    _tprintf( _T( "Supported functions:\n\n" ) );

    _tprintf( _T( "   load      Load Media\n" ) );
    _tprintf( _T( "   unload    Unload Media\n" ) );
    _tprintf( _T( "   lock      Lock Media\n" ) );
    _tprintf( _T( "   unlock    Unlock Media\n" ) );
    _tprintf( _T( "   rew       Rewind\n" ) );
    _tprintf( _T( "   run       Rewind Unload\n" ) );
    _tprintf( _T( "   read      Read Block  (0 = all    (default), 1-nnn = size)\n" ));
    _tprintf( _T( "   write     Write Block (0 = random (default), 1-nnn = size)\n" ));
    _tprintf( _T( "   wtm       Write [count] Tape Mark(s)\n" ) );
    _tprintf( _T( "   fsf       Forward Space [count] File(s)\n" ) );
    _tprintf( _T( "   fsr       Forward Space [count] Block(s)\n" ) );
    _tprintf( _T( "   bsf       Backspace [count] File(s)\n" ) );
    _tprintf( _T( "   bsr       Backspace [count] Block(s)\n" ) );
    _tprintf( _T( "   readpos   Read BlockId\n" ) );
    _tprintf( _T( "   locate    Locate <blockid>\n" ) );
    _tprintf( _T( "   erg       Erase Gap\n" ) );
    _tprintf( _T( "   comp      Set Compression [value] (0=off, 1=on (default))\n" ) );
    _tprintf( _T( "   ecc       Set ECC [value]         (0=off, 1=on (default))\n" ) );
    _tprintf( _T( "   margin    Set EOT Margin [value]  (0-nnn; default=0)\n" ));
    _tprintf( _T( "   status    Display status\n" ) );
    _tprintf( _T( "   ascii     Set ASCII data block display mode\n" ) );
    _tprintf( _T( "   ebcdic    Set EBCDIC data block display mode\n" ) );
    _tprintf( _T( "   dump      Display data block [offset[.length]]\n" ) );
    _tprintf( _T( "   erase     Erase entire tape (Data Security Erase)\n" ) );
    _tprintf( _T( "\n" ) );
    _tprintf( _T( "   help      Display usage information (what you're reading now)\n" ) );
    _tprintf( _T( "   ?         Synonym for 'help'\n" ) );
    _tprintf( _T( "   exit      Exit program\n" ) );
    _tprintf( _T( "   quit      Synonym for 'exit'\n" ) );
    _tprintf( _T( "\n" ) );
}

//////////////////////////////////////////////////////////////////////////////////////////
// Read a line from stdin...

CString ReadLine()
{
    CString strLine; char buff[80], *psz;
    do if ( NULL != ( psz = fgets( buff, NUM_CHARS( buff ), stdin ) ) ) strLine += psz;
    while ( strLine.IsEmpty() || _T( '\n' ) != strLine.Right(1)[0] ); return TrimBlanks(strLine);
}

//////////////////////////////////////////////////////////////////////////////////////////
// Forward declarations...

DWORD  w32_stape_open                 ();
DWORD  w32_stape_load                 ();
DWORD  w32_stape_unload               ();
DWORD  w32_stape_lock                 ();
DWORD  w32_stape_unlock               ();
DWORD  w32_stape_rewind               ();
DWORD  w32_stape_read                 ( DWORD dwBlkSize );
DWORD  w32_stape_write                ( DWORD dwBlkSize );
DWORD  w32_stape_write_tapemark       ( DWORD dwCount );
DWORD  w32_stape_erase_gap            ();
DWORD  w32_stape_erase_tape           ();
DWORD  w32_stape_set_margin           ( DWORD dwBytes );
DWORD  w32_stape_set_compression      ( BOOL bEnable );
DWORD  w32_stape_set_error_correction ( BOOL bEnable );
DWORD  w32_stape_update_status        ();
DWORD  w32_stape_report_status        ();

DWORD  w32_stape_skip_files           ( LONG lCount );
DWORD  w32_stape_skip_blocks          ( LONG lCount );

DWORD  w32_stape_read_abs_blockid     ( DWORD& dwAbsPos );
DWORD  w32_stape_locate_abs           ( DWORD  dwAbsPos );

//////////////////////////////////////////////////////////////////////////////////////////
// Global variables...

CString  g_strDeviceFileName;
CString  g_DeviceFilePrefix;    // _T( "\\\\.\\" )
CLocale  g_Locale;              // (from FishLib)

//////////////////////////////////////////////////////////////////////////////////////////

#define RANDOM_BLKSIZE_RANGE ( ( RANDOM_BLKSIZE_HIGH - RANDOM_BLKSIZE_LOW ) + 1 )

BYTE   g_bRandomByte     = 0;       // (random byte pattern)
DWORD  g_dwBufferSize    = 0;       // (current buffer size)
DWORD  g_dwActualIOBytes = 0;       // (#of bytes read or written)
BYTE*  g_pDataBuffer     = NULL;    // (random sized data buffer)
bool   g_bEBCDIC         = true;    // (data type in data buffer)

static inline BYTE gen_random_byte_pattern()
{
    return (BYTE) ( ( (unsigned) rand() ) % 256 );
}

//////////////////////////////////////////////////////////////////////////////////////////

BYTE* gen_random_sized_data_block()
{
    static bool bDidRandInit = false;

    if ( !bDidRandInit )
    {
        srand( (unsigned) time( NULL ) );
        bDidRandInit = true;
    }

    g_dwBufferSize = (DWORD) ( RANDOM_BLKSIZE_LOW + ( ((unsigned)rand()) % RANDOM_BLKSIZE_RANGE ) );
    ASSERT( RANDOM_BLKSIZE_LOW <= g_dwBufferSize && g_dwBufferSize <= RANDOM_BLKSIZE_HIGH );

    delete [] g_pDataBuffer;
    g_pDataBuffer = new BYTE [ g_dwBufferSize ];
    ASSERT( g_pDataBuffer );
    return g_pDataBuffer;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Support macro...

#define PRINT_ERROR( func, err )                      \
                                                      \
    _tprintf                                          \
    (                                                 \
        _T( "** %s: error; rc=%d (0x%08.8X): %s\n" )  \
                                                      \
        ,(func)                                       \
        ,(err)                                        \
        ,(err)                                        \
        ,g_Locale.FormatError( err )                  \
    )

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Lock for accessing the below information... The tape drive handle is used
// by both the monitor thread as well as our main ftape thread and thus must
// only be accessed with the lock held...

CRITICAL_SECTION                                          g_TapeStatusLock;
inline void LockTapeStatusData()   {EnterCriticalSection(&g_TapeStatusLock);}
inline void UnlockTapeStatusData() {LeaveCriticalSection(&g_TapeStatusLock);}

bool     g_bTapeIsMounted   = false;    // (set by media monitor worker thread)
HANDLE   g_hTapeDevice      = NULL;     // (handle to SCSI tape drive device)
HANDLE   g_hMonitorThread   = NULL;     // (handle to media monitor worker thread)
HANDLE   g_hShutdownEvent   = NULL;     // (event signaled to end worker thread)
HANDLE   g_hGetStatusEvent  = NULL;     // (wakes up worker thread to do its thing)
HANDLE   g_hGotStatusEvent  = NULL;     // (indicates worker thread did its thing)

// The monitor thread updates the above 'g_bTapeIsMounted' flag...

DWORD WINAPI MonitorMediaMountsThread( void* pArg );

//////////////////////////////////////////////////////////////////////////////////////////
// Worker thread to monitor media mounts/unmounts...

#define  STATUS_QUERY_FREQUENCY_SECS   (1)      // (maximum of once per second)
#define  SCSI_BUS_SETTLE_MILLISECS     (260)    // (see below) [*]

// [*] See Table 7 on page 35 of ANSI SCSI-2 standard (ANSI INCITS 131-1994 (R1999),
//     formerly ANSI X3.131-1994 (R1999)): The recommended SCSI "Reset to Selection
//     Delay" is defined at 250 msecs (milliseconds), which is the MAXIMUM time that
//     a SCSI target should wait after a bus reset until it's ready for reselection,
//     so waiting AT LEAST that long should guarantee the bus reset has completed.

DWORD  WINAPI  MonitorMediaMountsThread ( void* pArg )
{
    TRACE( _T("** MMMT: ENTRY...\n") );

    DWORD   dwRetCode;
    bool    bTapeWasMounted   = false;
    bool    bFirstTime        = true;

    HANDLE  hMultEventArray[2];

    hMultEventArray[0] = g_hShutdownEvent;      // first priority
    hMultEventArray[1] = g_hGetStatusEvent;     // second priority

    TRACE( _T("** main loop entry (await exit/work)...\n") );

    while
    (
        WAIT_OBJECT_0  !=  WaitForMultipleObjects
        (
            2,
            hMultEventArray,
            FALSE,
            STATUS_QUERY_FREQUENCY_SECS * 1000
        )
    )
    {
        TRACE( _T("** MMMT: awoken...\n") );

        /*
        **   Only retrieve the current status IF...
        **
        **        1.  We're specifically asked to do so...
        **                       -OR-
        **        2.  There's no tape currently mounted...
        **
        **   Otherwise, just go back to sleep. This prevents us
        **   from interfering with normal tape i/o operations....
        */

        TRACE( _T("** MMMT: wait for exit or work...\n") );

        LockTapeStatusData();

        if (1
            &&  g_bTapeIsMounted
            &&  !IsEventSet( g_hGetStatusEvent )
        )
        {
            // Tape is already mounted -AND- we have NOT
            // been specifically asked for the drive status...

            TRACE( _T("** MMMT: tape mounted && not asked for status; going back to sleep...\n") );

            UnlockTapeStatusData();
            continue;
        }

        // The tape is either NOT already mounted yet -OR- we have
        // been SPECIFICALLY ASKED to query the tape drive's status...

        TRACE( _T("** MMMT: tape NOT mounted ||  asked for status; working...\n") );

        ASSERT( !g_bTapeIsMounted || IsEventSet( g_hGetStatusEvent ) );

        //////////////////////////////////////////////////////////////////////
        //
        //  KB article Q111837: "INFO: ERROR_BUS_RESET May Be Benign"
        //
        //////////////////////////////////////////////////////////////////////
        //
        //  Tape API (application programming interface) functions
        //  may return an error code of ERROR_BUS_RESET when operating
        //  on SCSI tape devices. In many cases, you can ignore this
        //  error value and retry the operation. However, this error
        //  is fatal if received during a series of write operations
        //  because a tape drive cannot recover from a bus reset and
        //  continue writing.
        //
        //  When Windows boots up it resets the SCSI bus. This bus
        //  reset is reported by the tape drive in response to the
        //  first operation after the reset.
        //
        //  The code fragment shown below in the Sample Code section
        //  could be used to check for ERROR_BUS_RESET and clear it.
        //  The same technique could be used for other informational
        //  errors, such as ERROR_MEDIA_CHANGED, that may not be relevant
        //  at application startup.
        //
        //  Sample Code
        //
        //      ...
        //
        //      do {
        //          dwError = GetTapeStatus(hTape);
        //      } while (dwError == ERROR_BUS_RESET);
        //
        //      ...
        //
        //////////////////////////////////////////////////////////////////////

        TRACE( _T("** MMMT: retrieving status...\n") );

        while
        (1
            && !IsEventSet( g_hShutdownEvent )
            && g_hTapeDevice
            &&
            (
                ERROR_BUS_RESET      ==  (dwRetCode = GetTapeStatus( g_hTapeDevice ))
                ||
                ERROR_MEDIA_CHANGED  ==   dwRetCode
            )
        )
        {
            // Wait a little while for bus to settle...
            UnlockTapeStatusData();
            TRACE( _T("** MMMT: bus-reset or media-changed; sleeping a bit...\n") );
            Sleep( SCSI_BUS_SETTLE_MILLISECS );
            LockTapeStatusData();
        }

        if (IsEventSet( g_hShutdownEvent ))
        {
            UnlockTapeStatusData();
            break;
        }

        // Examine status...

        TRACE( _T("** MMMT: status retrieved; examining...\n") );

        if ( !g_hTapeDevice )
        {
            g_bTapeIsMounted = false;
        }
        else
        {
            switch ( dwRetCode )
            {
                case NO_ERROR:
                {
                    g_bTapeIsMounted = true;
                }
                break;

                case ERROR_NOT_READY:
                case ERROR_NO_MEDIA_IN_DRIVE:
                {
                    g_bTapeIsMounted = false;
                }
                break;

                default:
                {
                    ;   // nop (do nothing)
                }
                break;
            }
        }

        // Only update the tape status if it might have changed...

        if (0
            || (  g_bTapeIsMounted && !bTapeWasMounted )
            || ( !g_bTapeIsMounted &&  bTapeWasMounted )
            ||    bFirstTime
        )
        {
            g_bTapeIsMounted  =
              bTapeWasMounted =
                ( g_hTapeDevice && g_bTapeIsMounted );

            _tprintf
            (
                _T( "%s\n\tTape has been %s\n\n%s" )

                ,bFirstTime       ? _T(    ""    ) : _T(   "\n"      )
                ,g_bTapeIsMounted ? _T( "loaded" ) : _T( "unloaded"  )
                ,bFirstTime       ? _T(   "\n"   ) : _T( "Command? " )
            );

            bFirstTime = false;
        }

        // Tell caller current status is now available...

        TRACE( _T("** MMMT: SETting 'gOt status' event...\n") );
        VERIFY( SetEvent( g_hGotStatusEvent ) );

        // Go back to waiting for work...

        TRACE( _T("** MMMT: REsetting 'gEt status' event...\n") );
        VERIFY( ResetEvent( g_hGetStatusEvent ) );
        UnlockTapeStatusData();

        TRACE( _T("** MMMT: going back to sleep to await exit/work...\n") );
    }

    TRACE( _T("** MMMT: EXIT\n") );

    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_open()
{
    LockTapeStatusData();

    if ( !g_hTapeDevice )
    {
        g_hTapeDevice = CreateFile( g_DeviceFilePrefix + g_strDeviceFileName,
            GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );

        if ( INVALID_HANDLE_VALUE == g_hTapeDevice )
        {
            g_hTapeDevice = NULL;
            UnlockTapeStatusData();
            return GetLastError();
        }
    }

    UnlockTapeStatusData();

    return w32_stape_update_status();
}

//////////////////////////////////////////////////////////////////////////////////////////
// NOTE! if we can't GET the information, then we can't SET it either!.

bool                        g_bMediaParmsValid  = false;
TAPE_GET_MEDIA_PARAMETERS   g_GET_Media_parms   = {0};

bool                        g_bDriveParmsValid  = false;
TAPE_GET_DRIVE_PARAMETERS   g_GET_Drive_parms   = {0};

//////////////////////////////////////////////////////////////////////////////////////////
// Retrieve tape drive status (features / capabilities)...

DWORD w32_stape_update_status()
{
    memset( &g_GET_Drive_parms, 0, NUM_BYTES( g_GET_Drive_parms ) );
    memset( &g_GET_Media_parms, 0, NUM_BYTES( g_GET_Media_parms ) );

    g_bDriveParmsValid = false;
    g_bMediaParmsValid = false;

    // Ask the media monitor thread to take a sample...

    _tprintf( _T( "Retrieving status...\n" ) );

    ASSERT( g_hTapeDevice );                                // (required by thread)
    LockTapeStatusData();
    {
        TRACE( _T("** w32_stape_update_status: REsetting 'gOt status' event...\n") );
        TRACE( _T("** w32_stape_update_status: SETting 'gEt status' event...\n") );

        VERIFY( ResetEvent( g_hGotStatusEvent ) );          // (reset work event)
        VERIFY( SetEvent  ( g_hGetStatusEvent ) );          // (wake up thread)
    }
    UnlockTapeStatusData();

    TRACE( _T("** w32_stape_update_status: waiting for 'gOt status' event...\n") );
    WaitForSingleObject( g_hGotStatusEvent, INFINITE );     // (wait for status)

    // (Try to) Retrieve the media and drive information...

    DWORD  dwSize, dwRetCode;

    // There's no sense in trying to retrieve the MEDIA parameters
    // if there's no modea mounted in the drive!

    if ( g_bTapeIsMounted )
    {
        dwSize = NUM_BYTES(                                                         g_GET_Media_parms );
        g_bMediaParmsValid = ( NO_ERROR == ( dwRetCode =
            GetTapeParameters( g_hTapeDevice, GET_TAPE_MEDIA_INFORMATION, &dwSize, &g_GET_Media_parms ) ) )
            ? true : false;
    }

    dwSize = NUM_BYTES(                                                         g_GET_Drive_parms );
    g_bDriveParmsValid = ( NO_ERROR == ( dwRetCode =
        GetTapeParameters( g_hTapeDevice, GET_TAPE_DRIVE_INFORMATION, &dwSize, &g_GET_Drive_parms ) ) )
        ? true : false;

    dwRetCode = NO_ERROR;   // (ignore previous errors, if any)

    // (Try to) Request variable blocksized i/o...
    // (if we really NEED to and if we're ABLE to)

    if ( !g_bMediaParmsValid )
    {
        // If we were unable to GET the media information, then
        // it's quite likely that we can't try SETTING it either...

        _tprintf
        (
            _T( "Warning! Unable to request variable blocksized i/o\n" )
            _T( "         due to media parameters retrieval failure!\n" )
        );
    }
    else
    {
        // ---------------------------------------------------------------------
        //
        //                      PROGRAMMING NOTE
        //
        //
        // There are 3 separate flags dealing with variable blocksize mode:
        //
        //
        //     FeaturesLow   TAPE_DRIVE_VARIABLE_BLOCK   (0x00000800)
        //     FeaturesLow   TAPE_DRIVE_FIXED_BLOCK      (0x00000400)
        //     FeaturesHigh  TAPE_DRIVE_SET_BLOCK_SIZE   (0x80000010)
        //
        //
        // The first one (TAPE_DRIVE_VARIABLE_BLOCK) determines whether the
        // drive supports "variable-length block mode", whereas the third one
        // (TAPE_DRIVE_SET_BLOCK_SIZE) determines whether the drive supports
        // "setting the size of a fixed-length logical block or setting the
        // variable-length block mode".
        //
        // That is to say, it's possible for the drive to support variable-
        // length blocks but NOT support setting fixed-length blocks and thus
        // there's no need for the drive to support setting the blocksize at
        // all! That is, the drive supports variable-length blocks and ONLY
        // variable-length blocks, and thus TAPE_DRIVE_VARIABLE_BLOCK would
        // be TRUE whereas TAPE_DRIVE_SET_BLOCK_SIZE would be FALSE.
        //
        // Conversely, it's also possible for a drive to support multiple
        // different FIXED-length block sizes but NOT variable-length blocks.
        // In such a case TAPE_DRIVE_VARIABLE_BLOCK would be FALSE whereas
        // TAPE_DRIVE_SET_BLOCK_SIZE would be TRUE (in which case FeaturesLow
        // TAPE_DRIVE_FIXED_BLOCK should also be set)
        //
        // And of course the final case is a drive that supports a single
        // fixed-length blocksize, in which case TAPE_DRIVE_FIXED_BLOCK would
        // be TRUE but TAPE_DRIVE_VARIABLE_BLOCK and TAPE_DRIVE_SET_BLOCK_SIZE
        // would both be FALSE.
        //
        // THUS you need to take care to check ALL flags to determine whether
        // the drive supports your desired blocksize mode.
        //
        // Finally, if the drive DOES support variable-length block mode, you
        // should check the 'BlockSize' variable of the drive MEDIA parameters
        // to determine whether or not the drive happens to already be in the
        // desired mode. A 'BlockSize' value of zero means the drive is already
        // in variable-length block mode, whereas any other value indicates it
        // is in fixed-block mode.
        //
        // ---------------------------------------------------------------------

        bool bSetVarBlockIO = true;     // (what we MAY wish to do)

        if ( g_bDriveParmsValid )
        {
            if ( !( g_GET_Drive_parms.FeaturesLow & TAPE_DRIVE_VARIABLE_BLOCK ) )
            {
                // Drive does NOT appear to support variable blocksized i/o...

                // BUT if it is ALREADY set to do variable-sized blocks, then
                // that's okay, we can hopefully still used it. NOTE: this is
                // admittedly a highly unlikely (but nonethless conceivably
                // possible) case, and thus we SHOULD report it.

                if ( 0 == g_GET_Media_parms.BlockSize )
                {
                    bSetVarBlockIO = false;     // (probably DON'T need to do it)

                    _tprintf
                    (
                        _T( "Warning: Drive does NOT appear to support variable blocksized i/o,\n" )
                        _T( "BUT it appears to already be set in variable blocksized i/o mode!\n" )
                    );
                }
                else
                {
                    // Drive does NOT appear to support variable blocksized i/o, BUT if
                    // it supports SETTING the blocksize we'll TRY to set variable-length
                    // block mode anyway....    (but I highly doubt it'll work)

                    if ( g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_SET_BLOCK_SIZE )
                    {
                        bSetVarBlockIO = true;  // ( we'll try it but I doubt it'll work)

                        _tprintf
                        (
                            _T( "Warning: Drive does NOT appear to support variable blocksized i/o,\n" )
                            _T( "BUT it DOES report that it supports SETTING the blocksize, so we'll\n" )
                            _T( "TRY setting variable-blocksize mode...   (but I doubt it'll work!)\n" )
                        );
                    }
                    else
                    {
                        // Drive reports that not only does it not support variable blocksized
                        // i/o, but also that it's not already in variable blocksized AND does
                        // NOT support SETTING the blocksize either. This tape drive CANNOT be
                        // used with Hercules...

                        bSetVarBlockIO = false;     // (CAN'T do it)

                        _tprintf
                        (
                            _T( "ERROR: Drive reports that it not only does NOT support variable-sized\n" )
                            _T( "blocks, but ALSO that it does NOT support SETTING the blocksize either\n" )
                            _T( "(and it's not already SET to variable blocksized mode either!). Thus it\n" )
                            _T( "appears CERTAIN this tape drive CANNOT be used with Hercules. Sorry! :(\n" )
                        );
                    }
                }
            }
            else
            {
                // Drive reports that it DOES support variable-blocksized i/o, BUT...

                // We need to check to see whether or not it's already SET to variable-
                // blocksize i/o mode, and if it's not, whether or not it reports that
                // it's possible to set it to that mode...

                if ( 0 == g_GET_Media_parms.BlockSize )
                {
                    bSetVarBlockIO = false;     // (don't need to do it!)

                    _tprintf
                    (
                        _T( "Drive DOES appear to support variable blocksized i/o AND\n" )
                        _T( "appears to already be set to that mode too. (that's good!)\n" )
                    );
                }
                else
                {
                    // Drive reports it DOES support variable blocked i/o mode,
                    // but DOESN'T appear to be in that mode at the moment. We
                    // need to check to see if it's possible or not for us to
                    // REQUEST that it switch to variable-blocksize i/o mode...

                    if ( g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_SET_BLOCK_SIZE )
                    {
                        bSetVarBlockIO = false;     // (don't need to do it!)

                        _tprintf
                        (
                            _T( "Drive DOES appear to support variable blocksized i/o...\n" )
                            _T( "but does NOT appear to be currently set in that particular mode (yet).\n" )
                        );
                    }
                    else
                    {
                        bSetVarBlockIO = false;     // (CAN'T do it!)

                        _tprintf
                        (
                            _T( "ERROR: Drive reports it DOES support variable-blocksized i/o, BUT\n" )
                            _T( "that it not only is NOT already in that mode but ALSO that it does\n" )
                            _T( "NOT support SETTING it to that mode either! This is HIGHLY UNUSUAL!\n" )
                        );
                    }
                }
            }
        }
        else
        {
            // The drive parms are not available, so we can't know for certain
            // whether or not we CAN request variable blocksized i/o, but since
            // we WERE able to successfully retrieve the media information, we
            // will TRY requesting variable blocksized i/o anyway...

            _tprintf
            (
                _T( "Warning: Drive parameters unavailable. Unable to confidently\n" )
                _T( "         determine if drive supports (or supports setting of)\n" )
                _T( "         variable blocksized i/o mode...\n" )
            );

            // BUT BEFORE WE DO... let's check to see if maybe if maybe the drive
            // is already set in variable blocksized mode (by checking the media
            // parameters, if they're available). If the drive is already set to
            // variable blocksized i/o, then we don't need to request it. Else we
            // will TRY to request it...

            if ( 0 == g_GET_Media_parms.BlockSize )
            {
                _tprintf
                (
                    _T( "\n" )
                    _T( "However, it appears the drive is set in variable blocksized\n" )
                    _T( "i/o mode already, so I won't bother to try and request it.\n" )
                );
            }
            else
                bSetVarBlockIO = true;  // (otherwise try setting it anyway)
        }

        if ( bSetVarBlockIO )
        {
            _tprintf( _T( "\nRequesting variable blocksized i/o...\n\n" ) );

            TAPE_SET_MEDIA_PARAMETERS  SET_Media_parms;

            SET_Media_parms.BlockSize = 0;  // (zero means "variable")

            // (Note: the 'BlockSize' member is the ONLY member
            // in the above TAPE_SET_MEDIA_PARAMETERS structure)

            if ( NO_ERROR != ( dwRetCode = SetTapeParameters( g_hTapeDevice, SET_TAPE_MEDIA_INFORMATION, &SET_Media_parms ) ) )
            {
                _tprintf( _T( "** w32_stape_update_status: SetTapeParameters failed! rc=%d (0x%08.8X): %s\n\n" ),
                    dwRetCode, dwRetCode, g_Locale.FormatError( dwRetCode ) );

                _tprintf
                (
                    _T( "WARNING! The attempt to set the tape drive into variable blocksized i/o\n" )
                    _T( "mode FAILED! You MAY have problems using this tape drive with Hercules!\n" )
                );
            }
            else
            {
                // It appears to have worked...

                g_GET_Media_parms.BlockSize = 0;    // (because it worked!)

                _tprintf
                (
                    _T( "The attempt to set the tape drive into variable blocksized i/o mode appears\n" )
                    _T( "to have been successful. The drive should now perform variable blocksized i/o.\n" )
                );
            }
        }
    }

    _tprintf( _T( "\n" ) );

    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Report tape drive status (features / capabilities)...

DWORD w32_stape_report_status()
{
    DWORD  dwRetCode;

    if  ( !g_hTapeDevice )
    {
        if ( NO_ERROR != ( dwRetCode = w32_stape_open() ) )
            return dwRetCode;
    }

    w32_stape_update_status();

    _tprintf( _T( "Tape device:  %s\n"   ), g_strDeviceFileName );
    _tprintf( _T( "Status:       %s\n\n" ), g_bTapeIsMounted ? _T( "Ready (Loaded)" ) : _T( "Not Ready (Not Loaded)" ) );

    _tprintf( _T( "Drive information:\n\n" ) );

    if ( !g_bDriveParmsValid )
    {
        _tprintf( _T( "  (unavailable)\n" ) );
    }
    else
    {
        /*  TAPE_GET_DRIVE_PARAMETERS...

        BOOLEAN  ECC                    the device supports hardware error correction
        BOOLEAN  Compression            hardware data compression is enabled
        BOOLEAN  DataPadding            data padding is enabled (keeps tape streaming at a constant speed)
        BOOLEAN  ReportSetmarks         setmark reporting is enabled
        DWORD    DefaultBlockSize       Device's default fixed block size, in bytes.
        DWORD    MaximumBlockSize       Device's maximum block size, in bytes.
        DWORD    MinimumBlockSize       Device's minimum block size, in bytes.
        DWORD    MaximumPartitionCount  Maximum number of partitions that can be created on the device.
        DWORD    EOTWarningZoneSize     Indicates the number of bytes between the end-of-tape warning
                                        and the physical end of the tape.
        */

        _tprintf( _T( "  ECC                           %s\n" ),            g_GET_Drive_parms.ECC            ? _T( "Yes" ) : _T( "No" ) );
        _tprintf( _T( "  Compression                   %s\n" ),            g_GET_Drive_parms.Compression    ? _T( "Yes" ) : _T( "No" ) );
        _tprintf( _T( "  DataPadding                   %s\n" ),            g_GET_Drive_parms.DataPadding    ? _T( "Yes" ) : _T( "No" ) );
        _tprintf( _T( "  ReportSetmarks                %s\n" ),            g_GET_Drive_parms.ReportSetmarks ? _T( "Yes" ) : _T( "No" ) );

        _tprintf( _T( "\n" ) );

        _tprintf( _T( "  DefaultBlockSize              %s bytes\n" ),      g_Locale.FormatNumber( (__int64) g_GET_Drive_parms.DefaultBlockSize      ) );
        _tprintf( _T( "  MaximumBlockSize              %s bytes\n" ),      g_Locale.FormatNumber( (__int64) g_GET_Drive_parms.MaximumBlockSize      ) );
        _tprintf( _T( "  MinimumBlockSize              %s bytes\n" ),      g_Locale.FormatNumber( (__int64) g_GET_Drive_parms.MinimumBlockSize      ) );
        _tprintf( _T( "  MaximumPartitionCount         %s partitions\n" ), g_Locale.FormatNumber( (__int64) g_GET_Drive_parms.MaximumPartitionCount ) );
        _tprintf( _T( "  EOTWarningZoneSize            %s bytes\n" ),      g_Locale.FormatNumber( (__int64) g_GET_Drive_parms.EOTWarningZoneSize    ) );

        _tprintf( _T( "\n" ) );

        /*  TAPE_GET_DRIVE_PARAMETERS...

        DWORD    FeaturesHigh

        TAPE_DRIVE_ABS_BLK_IMMED     moves the tape to a device-specific block address
                                     and returns as soon as the move begins.
        TAPE_DRIVE_ABSOLUTE_BLK      moves the tape to a device specific block address.
        TAPE_DRIVE_END_OF_DATA       moves the tape to the end-of-data marker in a partition.
        TAPE_DRIVE_FILEMARKS         moves the tape forward (or backward) a specified number of filemarks.
        TAPE_DRIVE_LOAD_UNLOAD       enables and disables the device for further operations.
        TAPE_DRIVE_LOAD_UNLD_IMMED   supports immediate load and unload operations.
        TAPE_DRIVE_LOCK_UNLOCK       enables and disables the tape ejection mechanism.
        TAPE_DRIVE_LOCK_UNLK_IMMED   supports immediate lock and unlock operations.
        TAPE_DRIVE_LOG_BLK_IMMED     moves the tape to a logical block address in a partition
                                     and returns as soon as the move begins.
        TAPE_DRIVE_LOGICAL_BLK       moves the tape to a logical block address in a partition.
        TAPE_DRIVE_RELATIVE_BLKS     moves the tape forward (or backward) a specified number of blocks.
        TAPE_DRIVE_REVERSE_POSITION  moves the tape backward over blocks, filemarks, or setmarks.
        TAPE_DRIVE_REWIND_IMMEDIATE  supports immediate rewind operation.
        TAPE_DRIVE_SEQUENTIAL_FMKS   moves the tape forward (or backward) to the first occurrence
                                     of a specified number of consecutive filemarks.
        TAPE_DRIVE_SEQUENTIAL_SMKS   moves the tape forward (or backward) to the first occurrence
                                     of a specified number of consecutive setmarks.
        TAPE_DRIVE_SET_BLOCK_SIZE    supports setting the size of a fixed-length logical block
                                     or setting the variable-length block mode.
        TAPE_DRIVE_SET_COMPRESSION   enables and disables hardware data compression.
        TAPE_DRIVE_SET_ECC           enables and disables hardware error correction.
        TAPE_DRIVE_SET_PADDING       enables and disables data padding.
        TAPE_DRIVE_SET_REPORT_SMKS   enables and disables the reporting of setmarks.
        TAPE_DRIVE_SETMARKS          moves the tape forward (or reverse) a specified number of setmarks.
        TAPE_DRIVE_SPACE_IMMEDIATE   supports immediate spacing.
        TAPE_DRIVE_TENSION           supports tape tensioning.
        TAPE_DRIVE_TENSION_IMMED     supports immediate tape tensioning.
        TAPE_DRIVE_WRITE_FILEMARKS   writes filemarks.
        TAPE_DRIVE_WRITE_LONG_FMKS   writes long filemarks.
        TAPE_DRIVE_WRITE_MARK_IMMED  supports immediate writing of short and long filemarks.
        TAPE_DRIVE_WRITE_SETMARKS    writes setmarks.
        TAPE_DRIVE_WRITE_SHORT_FMKS  writes short filemarks.
        */

    #define PRINT_DRIVE_FEATURES( low_or_high, flag_name ) \
        _tprintf( _T( "  %-27.27s   %s\n" ), _T( #flag_name ), \
            ( g_GET_Drive_parms.low_or_high  &  flag_name ) ? _T( "Yes" ) : _T( "No" ) )

        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_ABS_BLK_IMMED     );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_ABSOLUTE_BLK      );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_END_OF_DATA       );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_FILEMARKS         );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_LOAD_UNLOAD       );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_LOAD_UNLD_IMMED   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_LOCK_UNLOCK       );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_LOCK_UNLK_IMMED   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_LOG_BLK_IMMED     );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_LOGICAL_BLK       );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_RELATIVE_BLKS     );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_REVERSE_POSITION  );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_REWIND_IMMEDIATE  );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SEQUENTIAL_FMKS   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SEQUENTIAL_SMKS   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SET_BLOCK_SIZE    );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SET_COMPRESSION   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SET_ECC           );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SET_PADDING       );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SET_REPORT_SMKS   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SETMARKS          );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_SPACE_IMMEDIATE   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_TENSION           );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_TENSION_IMMED     );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_WRITE_FILEMARKS   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_WRITE_LONG_FMKS   );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_WRITE_MARK_IMMED  );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_WRITE_SETMARKS    );
        PRINT_DRIVE_FEATURES ( FeaturesHigh, TAPE_DRIVE_WRITE_SHORT_FMKS  );

        _tprintf( _T( "\n" ) );

        /*  TAPE_GET_DRIVE_PARAMETERS...

        DWORD    FeaturesLow

        TAPE_DRIVE_COMPRESSION       supports hardware data compression.
        TAPE_DRIVE_CLEAN_REQUESTS    can report if cleaning is required.
        TAPE_DRIVE_ECC               supports hardware error correction.
        TAPE_DRIVE_EJECT_MEDIA       physically ejects the tape on a software eject.
        TAPE_DRIVE_ERASE_BOP_ONLY    performs the erase operation from the beginning-of-partition marker only.
        TAPE_DRIVE_ERASE_LONG        performs a long erase operation.
        TAPE_DRIVE_ERASE_IMMEDIATE   performs an immediate erase operation
                                     (returns when the erase operation begins)
        TAPE_DRIVE_ERASE_SHORT       performs a short erase operation.
        TAPE_DRIVE_FIXED             creates fixed data partitions.
        TAPE_DRIVE_FIXED_BLOCK       supports fixed-length block mode.
        TAPE_DRIVE_INITIATOR         creates initiator-defined partitions.
        TAPE_DRIVE_PADDING           supports data padding.
        TAPE_DRIVE_GET_ABSOLUTE_BLK  provides the current device-specific block address.
        TAPE_DRIVE_GET_LOGICAL_BLK   provides the current logical block address (and logical tape partition).
        TAPE_DRIVE_REPORT_SMKS       supports setmark reporting.
        TAPE_DRIVE_SELECT            creates select data partitions.
        TAPE_DRIVE_SET_EOT_WZ_SIZE   supports setting the end-of-medium warning size.
        TAPE_DRIVE_SET_CMP_BOP_ONLY  must be at the beginning of a partition before it can set compression on.
        TAPE_DRIVE_TAPE_CAPACITY     returns the maximum capacity of the tape.
        TAPE_DRIVE_TAPE_REMAINING    returns the remaining capacity of the tape.
        TAPE_DRIVE_VARIABLE_BLOCK    supports variable-length block mode.
        TAPE_DRIVE_WRITE_PROTECT     returns an error if the tape is write-enabled or write-protected.
        */

        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_COMPRESSION      );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_CLEAN_REQUESTS   );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_ECC              );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_EJECT_MEDIA      );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_ERASE_BOP_ONLY   );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_ERASE_LONG       );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_ERASE_IMMEDIATE  );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_ERASE_SHORT      );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_FIXED            );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_FIXED_BLOCK      );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_INITIATOR        );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_PADDING          );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_GET_ABSOLUTE_BLK );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_GET_LOGICAL_BLK  );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_REPORT_SMKS      );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_SELECT           );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_SET_EOT_WZ_SIZE  );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_SET_CMP_BOP_ONLY );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_TAPE_CAPACITY    );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_TAPE_REMAINING   );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_VARIABLE_BLOCK   );
        PRINT_DRIVE_FEATURES ( FeaturesLow, TAPE_DRIVE_WRITE_PROTECT    );
    }

    _tprintf( _T( "\n" ) );

    _tprintf( _T( "Media information:\n\n" ) );

    if ( !g_bMediaParmsValid )
    {
        _tprintf( _T( "  (unavailable)\n" ) );
    }
    else
    {
        /*  TAPE_GET_MEDIA_PARAMETERS...

        LARGE_INTEGER  Capacity        Total number of bytes on the current tape partition.
        LARGE_INTEGER  Remaining       Number of bytes between the current position and the
                                       end of the current tape partition.
        DWORD          BlockSize       Number of bytes per block.
        DWORD          PartitionCount  Number of partitions on the tape.
        BOOLEAN        WriteProtected  the tape is write-protected
        */

        LPCTSTR pszVariable = ( 0 == g_GET_Media_parms.BlockSize ) ? _T( " (variable)" ) : _T( " (fixed!)" );

        _tprintf( _T( "  Capacity                      %s bytes\n" ),      g_Locale.FormatNumber( (__int64) g_GET_Media_parms.Capacity.QuadPart  ) );
        _tprintf( _T( "  Remaining                     %s bytes\n" ),      g_Locale.FormatNumber( (__int64) g_GET_Media_parms.Remaining.QuadPart ) );
        _tprintf( _T( "  BlockSize                     %s bytes%s\n" ),    g_Locale.FormatNumber( (__int64) g_GET_Media_parms.BlockSize          ), pszVariable );
        _tprintf( _T( "  PartitionCount                %s partitions\n" ), g_Locale.FormatNumber( (__int64) g_GET_Media_parms.PartitionCount     ) );
        _tprintf( _T( "  WriteProtected                %s\n" ),            g_GET_Media_parms.WriteProtected ? _T( "Yes" ) : _T( "No" ) );
    }

    _tprintf( _T( "\nData block dump mode = %s\n" ), g_bEBCDIC ? _T( "EBCDIC" ) : _T( "ASCII" ) );

    return NO_ERROR;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_write_tapemark( DWORD dwCount )
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_WRITE_MARK_IMMED);

    DWORD dwRetCode;

    // NOTE: According to the documentation:
    //
    //    "A short filemark contains a short erase gap
    //     that cannot be overwritten unless the write
    //     operation is performed from the beginning of
    //     the partition or from an earlier long filemark."
    //
    //    "A long filemark contains a long erase gap that
    //     allows an application to position the tape at
    //     the beginning of the filemark and to overwrite
    //     the filemark and the erase gap."
    //
    // Thus we PREFER to write long filemarks, but if we can't,
    // we have no other choice but to default to "filemarks"...

    DWORD  dwTapemarkType = TAPE_LONG_FILEMARKS;

    if (0
        || !g_bDriveParmsValid
        || !( g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_WRITE_LONG_FMKS )
    )
    {
        dwTapemarkType = TAPE_FILEMARKS;
    }

    dwRetCode = WriteTapemark( g_hTapeDevice, dwTapemarkType, dwCount, bImmediate );

    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_set_margin( DWORD dwBytes )
{
    if ( !g_bDriveParmsValid )
    {
        _tprintf( _T( "Not possible  (drive parms unavailable)\n" ) );
        return NO_ERROR;
    }

    if ( !( g_GET_Drive_parms.FeaturesLow & TAPE_DRIVE_SET_EOT_WZ_SIZE ) )
    {
        _tprintf( _T( "Not supported\n" ) );
        return NO_ERROR;
    }

    DWORD  dwNewEOTWarningZoneSize = dwBytes;
    DWORD  dwRetCode;

    TAPE_SET_DRIVE_PARAMETERS   SET_Drive_parms;

    SET_Drive_parms.Compression        = g_GET_Drive_parms.Compression;
    SET_Drive_parms.DataPadding        = g_GET_Drive_parms.DataPadding;
    SET_Drive_parms.ECC                = g_GET_Drive_parms.ECC;
    SET_Drive_parms.EOTWarningZoneSize = dwNewEOTWarningZoneSize;
    SET_Drive_parms.ReportSetmarks     = g_GET_Drive_parms.ReportSetmarks;

    if ( NO_ERROR == ( dwRetCode = SetTapeParameters( g_hTapeDevice, SET_TAPE_DRIVE_INFORMATION, &SET_Drive_parms ) ) )
        g_GET_Drive_parms.EOTWarningZoneSize = dwNewEOTWarningZoneSize;

    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD  w32_stape_set_compression( BOOL bEnable )
{
    if ( !g_bDriveParmsValid )
    {
        _tprintf( _T( "Not possible  (drive parms unavailable)\n" ) );
        return NO_ERROR;
    }

    if ( !( g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_SET_COMPRESSION ) )
    {
        _tprintf( _T( "Not supported\n" ) );
        return NO_ERROR;
    }

    BOOL   bNewCompression = bEnable ? TRUE : FALSE;
    DWORD  dwRetCode;

    TAPE_SET_DRIVE_PARAMETERS  SET_Drive_parms;

    SET_Drive_parms.Compression        = bNewCompression;
    SET_Drive_parms.DataPadding        = g_GET_Drive_parms.DataPadding;
    SET_Drive_parms.ECC                = g_GET_Drive_parms.ECC;
    SET_Drive_parms.ReportSetmarks     = g_GET_Drive_parms.ReportSetmarks;
    SET_Drive_parms.EOTWarningZoneSize = g_GET_Drive_parms.EOTWarningZoneSize;

    if ( NO_ERROR == ( dwRetCode = SetTapeParameters( g_hTapeDevice, SET_TAPE_DRIVE_INFORMATION, &SET_Drive_parms ) ) )
        g_GET_Drive_parms.Compression = bNewCompression;

    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD  w32_stape_set_error_correction( BOOL bEnable )
{
    if ( !g_bDriveParmsValid )
    {
        _tprintf( _T( "Not possible  (drive parms unavailable)\n" ) );
        return NO_ERROR;
    }

    if ( !( g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_SET_ECC ) )
    {
        _tprintf( _T( "Not supported\n" ) );
        return NO_ERROR;
    }

    BOOL   bNewECC = bEnable ? TRUE : FALSE;
    DWORD  dwRetCode;

    TAPE_SET_DRIVE_PARAMETERS  SET_Drive_parms;

    SET_Drive_parms.Compression        = g_GET_Drive_parms.Compression;
    SET_Drive_parms.DataPadding        = g_GET_Drive_parms.DataPadding;
    SET_Drive_parms.ECC                = bNewECC;
    SET_Drive_parms.ReportSetmarks     = g_GET_Drive_parms.ReportSetmarks;
    SET_Drive_parms.EOTWarningZoneSize = g_GET_Drive_parms.EOTWarningZoneSize;

    if ( NO_ERROR == ( dwRetCode = SetTapeParameters( g_hTapeDevice, SET_TAPE_DRIVE_INFORMATION, &SET_Drive_parms ) ) )
        g_GET_Drive_parms.ECC = bNewECC;

    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_lock()
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_LOCK_UNLK_IMMED);

    DWORD dwRetCode;
    dwRetCode = PrepareTape( g_hTapeDevice, TAPE_LOCK, bImmediate );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_unlock()
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_LOCK_UNLK_IMMED);

    DWORD dwRetCode;
    dwRetCode = PrepareTape( g_hTapeDevice, TAPE_UNLOCK, bImmediate );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_load()
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_LOAD_UNLD_IMMED);

    DWORD dwRetCode;
    dwRetCode = PrepareTape( g_hTapeDevice, TAPE_LOAD, bImmediate );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_unload()
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_LOAD_UNLD_IMMED);

    DWORD dwRetCode;
    dwRetCode = PrepareTape( g_hTapeDevice, TAPE_UNLOAD, bImmediate );
    LockTapeStatusData();
    {
        TRACE( _T("** w32_stape_unload: REsetting 'gOt status' event...\n") );
        TRACE( _T("** w32_stape_unload: SETting 'gEt status' event...\n") );

        VERIFY( ResetEvent( g_hGotStatusEvent ) );      // (reset work event)
        VERIFY( SetEvent  ( g_hGetStatusEvent ) );      // (wake up thread)
    }
    UnlockTapeStatusData();
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_rewind()
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_REWIND_IMMEDIATE);

    DWORD dwRetCode;
    dwRetCode = SetTapePosition( g_hTapeDevice, TAPE_REWIND, 0, 0, 0, bImmediate );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_erase_gap()
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesLow & TAPE_DRIVE_ERASE_IMMEDIATE);

    DWORD dwRetCode;
    dwRetCode = EraseTape( g_hTapeDevice, TAPE_ERASE_SHORT, bImmediate );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_erase_tape()
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesLow & TAPE_DRIVE_ERASE_IMMEDIATE);

    for (;;)
    {
        cout << _T( "Erase entire tape?! (y/n): " );
        CString str; str = TrimBlanks(ReadLine());
        if ( str.IsEmpty() ) continue; str = str.Left(1); str.MakeLower();
        if ( str == _T( 'y' ) ) break; if ( str != _T( 'n' ) ) continue;
        return ERROR_CANCELLED;
    }

    DWORD dwRetCode = EraseTape( g_hTapeDevice, TAPE_ERASE_LONG, bImmediate );

    if ( bImmediate && NO_ERROR == dwRetCode )
        cout << _T( "Erase in progress...\n" );

    if ( !bImmediate && NO_ERROR == dwRetCode )
        cout << _T( "Erase complete.\n" );

    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_skip_files( LONG lCount )
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_SPACE_IMMEDIATE);

    LARGE_INTEGER  liCount;  liCount.QuadPart = lCount;

    DWORD dwRetCode;
    dwRetCode = SetTapePosition( g_hTapeDevice, TAPE_SPACE_FILEMARKS, 0,
        liCount.LowPart, liCount.HighPart, bImmediate );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_skip_blocks( LONG lCount )
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_SPACE_IMMEDIATE);

    LARGE_INTEGER  liCount;  liCount.QuadPart = lCount;

    DWORD dwRetCode;
    dwRetCode = SetTapePosition( g_hTapeDevice, TAPE_SPACE_RELATIVE_BLOCKS, 0,
        liCount.LowPart, liCount.HighPart, bImmediate );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_read_abs_blockid( DWORD& dwAbsPos )
{
    DWORD dwDummyPart, dwDummyPosHigh;

    DWORD dwRetCode = GetTapePosition
    (
        g_hTapeDevice,
        TAPE_ABSOLUTE_POSITION,
        &dwDummyPart,
        &dwAbsPos,
        &dwDummyPosHigh
    );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD w32_stape_locate_abs( DWORD dwAbsPos )
{
    BOOL bImmediate = g_bDriveParmsValid &&
        (g_GET_Drive_parms.FeaturesHigh & TAPE_DRIVE_ABS_BLK_IMMED);

    DWORD dwRetCode = SetTapePosition
    (
        g_hTapeDevice,
        TAPE_ABSOLUTE_BLOCK,
        0,
        dwAbsPos,
        0,
        bImmediate
    );
    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD  w32_stape_read( DWORD dwBlkSize )
{
    if ( !( g_dwBufferSize = dwBlkSize ) )
        g_dwBufferSize = RANDOM_BLKSIZE_HIGH;

    delete [] g_pDataBuffer;

    g_pDataBuffer = new BYTE [ g_dwBufferSize ]; ASSERT( g_pDataBuffer );

    if ( !g_pDataBuffer )
        return ERROR_NOT_ENOUGH_MEMORY;

    g_dwActualIOBytes = 0;

    BOOL bSuccess = ReadFile( g_hTapeDevice, g_pDataBuffer, g_dwBufferSize, &g_dwActualIOBytes, NULL );
    DWORD dwLastError = GetLastError();

    if ( !bSuccess )
    {
        _tprintf( _T( "ERROR; rc=%d (0x%08.8X): %s\n" ),
            dwLastError, dwLastError, g_Locale.FormatError( dwLastError ) );
    }

    _tprintf
    (
        _T( "Read %s of %s bytes.%s\n" )

        ,g_Locale.FormatNumber( (__int64) g_dwActualIOBytes )
        ,g_Locale.FormatNumber( (__int64) g_dwBufferSize )

        ,bSuccess && g_dwActualIOBytes == g_dwBufferSize && ERROR_MORE_DATA == dwLastError ?
            CString( _T( " (" ) ) + g_Locale.FormatError( ERROR_MORE_DATA ) + CString( _T( ")" ) )
            : _T( "" )
    );

    return NO_ERROR;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD  w32_stape_write( DWORD dwBlkSize )
{
    if ( dwBlkSize )
    {
        delete [] g_pDataBuffer;
        g_pDataBuffer = new BYTE [ g_dwBufferSize = dwBlkSize ];
    }
    else
    {
        gen_random_sized_data_block();
        ASSERT( g_dwBufferSize <= RANDOM_BLKSIZE_HIGH );
    }

    if ( !g_pDataBuffer )
        return ERROR_NOT_ENOUGH_MEMORY;

    g_bRandomByte = gen_random_byte_pattern();
    memset( g_pDataBuffer, g_bRandomByte, g_dwBufferSize );

    g_dwActualIOBytes = 0;

    BOOL bSuccess = WriteFile( g_hTapeDevice, g_pDataBuffer, g_dwBufferSize, &g_dwActualIOBytes, NULL );
    DWORD dwLastError = GetLastError();

    if ( !bSuccess )
    {
        _tprintf( _T( "ERROR; rc=%d (0x%08.8X): %s\n" ),
            dwLastError, dwLastError, g_Locale.FormatError( dwLastError ) );
    }

    _tprintf
    (
        _T( "Wrote %s of %s bytes.\n" )

        ,g_Locale.FormatNumber( (__int64) g_dwActualIOBytes )
        ,g_Locale.FormatNumber( (__int64) g_dwBufferSize )
    );

    return NO_ERROR;
}

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

volatile  BOOL  g_bCtrl_C_Flag  = FALSE;      // (used to abort 'dump' command)

DWORD ftape_main( int argc, TCHAR** argv )
{
    g_DeviceFilePrefix  = _T( TAPE_DEVICE_PREFIX );

    if ( g_strDeviceFileName.IsEmpty() )
        g_strDeviceFileName = _T( DEFAULT_TAPE_DEVICE );

    static CString strPrevDeviceFileName;

    // Parse command line...

    CString strArg, strCommand;
    long lCount = -1;
    static const long lMinCount = LONG_MIN;
    static const long lMaxCount = LONG_MAX;
    bool bGotCount = false;
    TCHAR* pszStopChar;
    long lDumpOff = -1;
    long lDumpLen = -1;
    bool bGotDumpArg = false;

    for ( int i=1; i < argc; i++ )
    {
        strArg = argv[i];

        if ( strArg.Left(1) == _T( '-' ) )
        {
            if ( strArg.CollateNoCase( _T( "-f" ) ) != 0
                || (i+1) >= argc )
            {
                usage();
                return NO_ERROR;
            }

            g_strDeviceFileName = argv[++i];
        }
        else
        {
            if ( strCommand.IsEmpty() )
            {
                strCommand = strArg;
            }
            else
            {
                if ( strCommand.CollateNoCase( _T( "dump" ) ) != 0 )
                {
                    lCount = _tcstol( strArg, &pszStopChar, 0 );

                    if (0
                        || bGotCount
                        ||  (                    0 == lCount                 && EINVAL == errno )
                        ||  ( ( lMinCount == lCount || lMaxCount == lCount ) && ERANGE == errno )
                        || !( 0 == *pszStopChar || _istspace( *pszStopChar ) )
                    )
                    {
                        usage();
                        return NO_ERROR;
                    }

                    bGotCount = true;
                }
                else // "dump length" or "dump offset.length"
                {
                    if (bGotDumpArg)
                    {
                        usage();
                        return NO_ERROR;
                    }
                    else
                    {
                        lDumpOff = _tcstol( strArg, &pszStopChar, 0 );

                        if (0
                            ||  (                       0 == lDumpOff                && EINVAL == errno )
                            ||  ( ( lMinCount == lDumpOff || lMaxCount == lDumpOff ) && ERANGE == errno )
                            || !( 0 == *pszStopChar || _T( '.' ) == *pszStopChar )
                            ||  ( lDumpOff < 0 )
                            ||  (      0    == *pszStopChar && lDumpOff >  (long)g_dwActualIOBytes ) // (is a length)
                            ||  ( _T( '.' ) == *pszStopChar && lDumpOff >= (long)g_dwActualIOBytes ) // (is an offset)
                        )
                        {
                            usage();
                            return NO_ERROR;
                        }

                        if ( _T( '.' ) == *pszStopChar )
                        {
                            lDumpLen = _tcstol( pszStopChar+1, &pszStopChar, 0 );

                            if (0
                                ||  (                       0 == lDumpLen                && EINVAL == errno )
                                ||  ( ( lMinCount == lDumpLen || lMaxCount == lDumpLen ) && ERANGE == errno )
                                || !( 0 == *pszStopChar )
                                ||  ( lDumpLen < 0 )
                                ||  ( (lDumpOff + lDumpLen) > (long)g_dwActualIOBytes )
                            )
                            {
                                usage();
                                return NO_ERROR;
                            }
                        }
                        else
                        {
                            lDumpLen = lDumpOff;
                            lDumpOff = 0;
                        }

                        bGotDumpArg = true;
                    }
                }
            }
        }
    }

    if (0
        || strCommand == _T( "?" )
        || strCommand.CompareNoCase( _T( "help" ) ) == 0
    )
    {
        usage();
        return NO_ERROR;
    }

    ///////////////////////////////////////////////////
    // Open the tape file if it's not already open...
    // (close the previous one first if it's opened)
    ///////////////////////////////////////////////////

    CString  strFunc;
    DWORD    dwRetCode;

    if ( g_hTapeDevice )
    {
        if ( g_strDeviceFileName.CompareNoCase( strPrevDeviceFileName ) != 0 )
        {
            LockTapeStatusData();
            {
                VERIFY( CloseHandle( g_hTapeDevice ) );
                g_hTapeDevice = NULL;
            }
            UnlockTapeStatusData();
        }
    }

    if ( !g_hTapeDevice )
    {
        strPrevDeviceFileName = g_strDeviceFileName;

        if ( NO_ERROR != ( dwRetCode = w32_stape_open() ) )
        {
            PRINT_ERROR( _T( "w32_stape_open" ), dwRetCode );
            return dwRetCode;
        }
    }

    //////////////  commands that default to 'on' /////////////////

    if ( strCommand.CollateNoCase( _T( "comp" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 1;

        strFunc   = _T( "w32_stape_set_compression" );
        dwRetCode =      w32_stape_set_compression( lCount );
    }

    else

    if ( strCommand.CollateNoCase( _T( "ecc" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 1;

        strFunc   = _T( "w32_stape_set_error_correction" );
        dwRetCode =      w32_stape_set_error_correction( lCount );
    }

    else

    //////////////  commands that default to zero  /////////////////

    if ( strCommand.CollateNoCase( _T( "read" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 0;

        strFunc   = _T( "w32_stape_read" );
        dwRetCode =      w32_stape_read( lCount );
    }

    else

    if ( strCommand.CollateNoCase( _T( "write" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 0;

        strFunc   = _T( "w32_stape_write" );
        dwRetCode =      w32_stape_write( lCount );
    }

    else

    if ( strCommand.CollateNoCase( _T( "margin" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 0;

        strFunc   = _T( "w32_stape_set_margin" );
        dwRetCode =      w32_stape_set_margin( lCount );
    }

    else

    //////////////  commands that default to one   /////////////////

    if ( strCommand.CollateNoCase( _T( "wtm" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 1;
        if ( lCount <= 0 )
        {
            usage();
            return NO_ERROR;
        }

        strFunc   = _T( "w32_stape_write_tapemark" );
        dwRetCode =      w32_stape_write_tapemark( lCount );
    }

    else

    if ( strCommand.CollateNoCase( _T( "fsf" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 1;
        if ( lCount <= 0 )
        {
            usage();
            return NO_ERROR;
        }

        strFunc   = _T( "w32_stape_skip_files" );
        dwRetCode =      w32_stape_skip_files( lCount );
    }

    else

    if ( strCommand.CollateNoCase( _T( "bsf" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 1;
        if ( lCount <= 0 )
        {
            usage();
            return NO_ERROR;
        }

        strFunc   = _T( "w32_stape_skip_files" );
        dwRetCode =      w32_stape_skip_files( -lCount );
    }

    else

    if ( strCommand.CollateNoCase( _T( "fsr" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 1;
        if ( lCount <= 0 )
        {
            usage();
            return NO_ERROR;
        }

        strFunc   = _T( "w32_stape_skip_blocks" );
        dwRetCode =      w32_stape_skip_blocks( lCount );
    }

    else

    if ( strCommand.CollateNoCase( _T( "bsr" ) ) == 0 )
    {
        if ( !bGotCount ) lCount = 1;
        if ( lCount <= 0 )
        {
            usage();
            return NO_ERROR;
        }

        strFunc   = _T( "w32_stape_skip_blocks" );
        dwRetCode =      w32_stape_skip_blocks( -lCount );
    }

    else

    ///////////////////  special  ///////////////////

    if ( strCommand.CollateNoCase( _T( "readpos" ) ) == 0 )
    {
        if ( bGotCount )
        {
            usage();
            return NO_ERROR;
        }

        DWORD  dwAbsPos = 0;

        strFunc   = _T( "w32_stape_read_abs_blockid" );
        dwRetCode =      w32_stape_read_abs_blockid( dwAbsPos );

        if ( NO_ERROR == dwRetCode )
            _tprintf( _T( "\nPos = 0x%08.8X\n\n" ) ,dwAbsPos );
    }

    else

    if ( strCommand.CollateNoCase( _T( "locate" ) ) == 0 )
    {
        if ( !bGotCount )
        {
            usage();
            return NO_ERROR;
        }

        strFunc   = _T( "w32_stape_locate_abs" );
        dwRetCode =      w32_stape_locate_abs( lCount );
    }

    else

    if ( strCommand.CollateNoCase( _T( "dump" ) ) == 0 )
    {
        if ( g_pDataBuffer && g_dwActualIOBytes )
        {
            BYTE*   pData   = g_pDataBuffer;
            size_t  offset  = 0;
            size_t  length  = g_dwActualIOBytes;

            if ( bGotDumpArg )
            {
                offset = lDumpOff;
                length = lDumpLen;
                pData += offset;
            }

            ASSERT // (sanity check)
            (1
                && ( pData +    0   ) >= ( g_pDataBuffer +         0         )
                && ( pData +    0   ) <  ( g_pDataBuffer + g_dwActualIOBytes )
                && ( pData + length ) <= ( g_pDataBuffer + g_dwActualIOBytes )
            );

            // Dump piece-meal in case they want to cancel it...

            size_t  remaining  = length;
            size_t  chunk_size = 0x100;

            g_bCtrl_C_Flag = FALSE;

            do
            {
                if (chunk_size > remaining)
                    chunk_size = remaining;

                _tprintf( _T( "%s" ),
                    FormatHexDump( pData, chunk_size, g_bEBCDIC, offset, true ) );

                pData     += chunk_size;
                offset    += chunk_size;
                remaining -= chunk_size;
            }
            while ( remaining && !g_bCtrl_C_Flag );

            if ( g_bCtrl_C_Flag )
                _tprintf( _T( "\n(canceled)\n" ) );
    
            g_bCtrl_C_Flag = FALSE;

            dwRetCode = NO_ERROR;
        }
        else
        {
            _tprintf( _T( "You need to issue a read or write command first\n" ) );
            _tprintf( _T( "so a buffer gets created that you can then dump.\n" ) );
        }
    }

    else
    {
        //////////////  commands with no paramaters    /////////////////

        if ( bGotCount )
        {
            usage();
            return NO_ERROR;
        }

        else

        if ( strCommand.CollateNoCase( _T( "load" ) ) == 0 )
        {
            strFunc   = _T( "w32_stape_load" );
            dwRetCode =      w32_stape_load();
        }

        else

        if ( strCommand.CollateNoCase( _T( "unload" ) ) == 0 )
        {
            strFunc   = _T( "w32_stape_unload" );
            dwRetCode =      w32_stape_unload();
        }

        else

        if ( strCommand.CollateNoCase( _T( "lock" ) ) == 0 )
        {
            strFunc   = _T( "w32_stape_lock" );
            dwRetCode =      w32_stape_lock();
        }

        else

        if ( strCommand.CollateNoCase( _T( "unlock" ) ) == 0 )
        {
            strFunc   = _T( "w32_stape_unlock" );
            dwRetCode =      w32_stape_unlock();
        }

        else

        if ( strCommand.CollateNoCase( _T( "rew" ) ) == 0 )
        {
            strFunc   = _T( "w32_stape_rewind" );
            dwRetCode =      w32_stape_rewind();
        }

        else

        if ( strCommand.CollateNoCase( _T( "run" ) ) == 0 )
        {
            strFunc                     = _T( "w32_stape_rewind" );
            if ( NO_ERROR == (dwRetCode =      w32_stape_rewind()) )
            {
                strFunc   = _T( "w32_stape_unload" );
                dwRetCode =      w32_stape_unload();
            }
        }

        else

        if ( strCommand.CollateNoCase( _T( "erg" ) ) == 0 )
        {
            strFunc   = _T( "w32_stape_erase_gap" );
            dwRetCode =      w32_stape_erase_gap();
        }

        else

        if ( strCommand.CollateNoCase( _T( "erase" ) ) == 0 )
        {
            strFunc   = _T( "w32_stape_erase_tape" );
            dwRetCode =      w32_stape_erase_tape();
        }

        else

        if ( strCommand.CollateNoCase( _T( "status" ) ) == 0 )
        {
            strFunc   = _T( "w32_stape_report_status" );
            dwRetCode =      w32_stape_report_status();
        }

        else

        if ( strCommand.CollateNoCase( _T( "ascii" ) ) == 0 )
        {
            g_bEBCDIC = false;
            dwRetCode = NO_ERROR;
        }

        else

        if ( strCommand.CollateNoCase( _T( "ebcdic" ) ) == 0 )
        {
            g_bEBCDIC = true;
            dwRetCode = NO_ERROR;
        }

        else

        {
            usage();
            return NO_ERROR;
        }
    }

    // If something went wrong, show the error...

    if ( NO_ERROR != dwRetCode )
        PRINT_ERROR( strFunc, dwRetCode );

    return dwRetCode;
}

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

#define  MAX_CMDLINE_LEN   256
#define  MAX_ARGS          10

int     new_argc;
char    command_string [ MAX_CMDLINE_LEN ];
TCHAR*  argv_array     [    MAX_ARGS     ];

//-------------------------------------------------------------------
// Subroutine to parse an argument string. The string that is passed
// is modified in-place by inserting null characters at the end of
// each argument found. The returned array of argument pointers
// then points to each argument found in the original string. Any
// argument that begins with '#' comment indicator causes early
// termination of the parsing and is not included in the count. Any
// argument found that starts with a double-quote character causes
// all characters following the double-quote up to the next double-
// quote to be included as part of that argument. The quotes them-
// selves are not considered part of any argument and are ignored.
// p            Points to string to be parsed.
// maxargc      Maximum allowable number of arguments. (Prevents
//              overflowing the pargv array)
// pargv        Pointer to buffer for argument pointer array.
// pargc        Pointer to number of arguments integer result.
// Returns number of arguments found. (same value as at *pargc)
//-------------------------------------------------------------------

int parse_args ( TCHAR* p, int maxargc, TCHAR** pargv, int* pargc )
{
    for ( *pargc=0; *pargc < maxargc; ++*pargc ) pargv[*pargc] = NULL;

    *pargc = 0;
    *pargv = NULL;

    while (*p && *pargc < maxargc)
    {
        while (*p && isspace(*p)) p++; if (!*p) break; // find start of arg

        if (*p == _T('#')) break; // stop on comments

        *pargv = p; ++*pargc; // count new arg

        while (*p && !isspace(*p) && *p != _T('\"')) p++; if (!*p) break; // find end of arg

        if (*p == _T('\"'))
        {
            if (p == *pargv) *pargv = p+1;
            while (*++p && *p != _T('\"')); if (!*p) break; // find end of quoted string
        }

        *p++ = 0; // mark end of arg
        pargv++; // next arg ptr
    }

    return *pargc;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Ctrl+C handler...

BOOL ConsoleCtrlHandler( DWORD dwSignalType )
{
    if ( CTRL_C_EVENT == dwSignalType )
    {
        g_bCtrl_C_Flag = TRUE;
        return TRUE;
    }

    return FALSE;
}

//////////////////////////////////////////////////////////////////////////////////////////

int _tmain( int argc, TCHAR** argv, TCHAR** envp )
{
    Copyright();

    // Initialize...

    if (0
        || !AfxWinInit( GetModuleHandle( NULL ), NULL, GetCommandLine(), 0 )
        || !SetConsoleCtrlHandler( (PHANDLER_ROUTINE) ConsoleCtrlHandler, TRUE )
    )
        return 99;

    InitializeCriticalSection( &g_TapeStatusLock );

    g_hShutdownEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    ASSERT( g_hShutdownEvent ); if ( !g_hShutdownEvent ) return 99;

    g_hGetStatusEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    ASSERT( g_hGetStatusEvent ); if ( !g_hGetStatusEvent ) return 99;

    g_hGotStatusEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    ASSERT( g_hGotStatusEvent ); if ( !g_hGotStatusEvent ) return 99;

    // Create worker thread that will monitor for media changes...

    DWORD dwDummyThreadId = 0;      // (we don't need it)
    g_hMonitorThread = CreateThread
    (
        NULL,                       // pointer to security attributes
        0,                          // initial thread stack size
        MonitorMediaMountsThread,   // pointer to thread function
        NULL,                       // argument for new thread
        0,                          // creation flags
        &dwDummyThreadId            // pointer to receive thread ID
    );
    ASSERT( g_hMonitorThread ); if ( !g_hMonitorThread ) return 99;

    // Process command-line...

    _tcsncpy( command_string, GetCommandLine(), MAX_CMDLINE_LEN );
    command_string[ MAX_CMDLINE_LEN-1 ] = 0;
    parse_args( command_string, MAX_ARGS-1, argv_array+1, &new_argc );
    new_argc--;

    // Loop until told to exit...

    bool bFirstTime = true;

    while (!g_bCtrl_C_Flag)
    {
        if ( bFirstTime || new_argc )
        {
            new_argc++; argv_array[0] = argv[0];
            ftape_main( new_argc, argv_array );
            cout << _T("\n");
        }

        bFirstTime = false;

        cout << _T("Command? ");
        command_string[0] = 0;
        cin.getline( command_string, NUM_BYTES( command_string ) );
        command_string[ NUM_BYTES( command_string ) - 1 ] = 0;
        cout << _T("\n");

        if (!command_string[0])
            break;

        if ( parse_args( command_string, MAX_ARGS-1, argv_array+1, &new_argc ) == 1
            && (_tcsicoll( _T( "exit" ), argv_array[1] ) == 0 ||
                _tcsicoll( _T( "quit" ), argv_array[1] ) == 0))
            break;
    }

    // Shut everything down...

    if ( !g_bTapeIsMounted )
    {
        cout << _T( "\nExiting...\n" );
    }

    // Tell worker thread to shutdown...

    VERIFY( SetEvent( g_hShutdownEvent ) );

    // Wait for worker thread to shutdown...

    WaitForSingleObject( g_hMonitorThread, INFINITE );

    // Close handles and exit...

    VERIFY( CloseHandle( g_hMonitorThread ) );
    g_hMonitorThread = NULL;

    VERIFY( CloseHandle( g_hShutdownEvent ) );
    g_hShutdownEvent = NULL;

    VERIFY( CloseHandle( g_hGetStatusEvent ) );
    g_hGetStatusEvent = NULL;

    VERIFY( CloseHandle( g_hGotStatusEvent ) );
    g_hGotStatusEvent = NULL;

    if ( g_hTapeDevice )
    {
        VERIFY( CloseHandle( g_hTapeDevice ) );
        g_hTapeDevice = NULL;
    }

    DeleteCriticalSection( &g_TapeStatusLock );

    delete [] g_pDataBuffer;

#ifdef _DEBUG
    _tprintf( _T( "\nPress any key to continue" ) );
    while (!_kbhit()) Sleep(100); _tprintf( _T( "\n" ) );
#endif

    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
